import structlog
from typing import Generic, TypeVar
from pydantic import BaseModel, ConfigDict
from rid_lib.types import KoiNetNode

from .secure import PrivateKey, PublicKey
from .api_models import RequestModels, ResponseModels

log = structlog.stdlib.get_logger()


T = TypeVar("T", bound=RequestModels | ResponseModels)

class SignedEnvelope(BaseModel, Generic[T]):
    payload: T
    source_node: KoiNetNode
    target_node: KoiNetNode
    signature: str
    
    model_config = ConfigDict(exclude_none=True)
    
    def verify_with(self, pub_key: PublicKey):
        """Verifies signed envelope with public key.
        
        Raises `cryptography.exceptions.InvalidSignature` on failure.
        """
        
        # IMPORTANT: calling `model_dump()` loses all typing! when converting between SignedEnvelope and UnsignedEnvelope, use the Pydantic classes, not the dictionary form
        
        unsigned_envelope = UnsignedEnvelope[T](
            payload=self.payload,
            source_node=self.source_node,
            target_node=self.target_node 
        )
        
        log.debug(f"Verifying envelope: {unsigned_envelope.model_dump_json(exclude_none=True)}")

        pub_key.verify(
            self.signature,
            unsigned_envelope.model_dump_json(exclude_none=True).encode()
        )

class UnsignedEnvelope(BaseModel, Generic[T]):
    payload: T
    source_node: KoiNetNode
    target_node: KoiNetNode
    
    model_config = ConfigDict(exclude_none=True)
    
    def sign_with(self, priv_key: PrivateKey) -> SignedEnvelope[T]:
        """Signs with private key and returns `SignedEnvelope`."""
        
        log.debug(f"Signing envelope: {self.model_dump_json(exclude_none=True)}")
        log.debug(f"Type: [{type(self.payload)}]")
        
        signature = priv_key.sign(
            self.model_dump_json(exclude_none=True).encode()
        )
        
        return SignedEnvelope(
            payload=self.payload,
            source_node=self.source_node,
            target_node=self.target_node,
            signature=signature
        )



from rid_lib.ext import Bundle
from ..identity import NodeIdentity
from ..processor.kobj_queue import KobjQueue


class ProfileMonitor:
    """Processes changes to node profile in the config."""
    def __init__(
        self,
        kobj_queue: KobjQueue,
        identity: NodeIdentity
    ):
        self.kobj_queue = kobj_queue
        self.identity = identity
        
    def start(self):
        """Processes identity bundle generated from config."""
        self_bundle = Bundle.generate(
            rid=self.identity.rid,
            contents=self.identity.profile.model_dump()
        )
        
        self.kobj_queue.push(bundle=self_bundle)


import structlog
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from ..network.graph import NetworkGraph
from ..config.base import BaseNodeConfig
from ..identity import NodeIdentity
from ..network.event_queue import EventQueue
from ..protocol.event import Event, EventType

log = structlog.stdlib.get_logger()


class Handshaker:
    """Handles handshaking with other nodes."""
    def __init__(
        self, 
        cache: Cache, 
        identity: NodeIdentity, 
        event_queue: EventQueue,
        config: BaseNodeConfig,
        graph: NetworkGraph
    ):
        self.config = config
        self.cache = cache
        self.identity = identity
        self.event_queue = event_queue
        self.graph = graph
        
    def start(self):
        """Attempts handshake with first contact on startup.
        
        Handshake occurs if first contact is set in the config, the first
        contact is not already known to this node, and this node does not
        already have incoming edges with node providers.
        """
        if not self.config.koi_net.first_contact.rid:
            return
        
        if self.cache.read(self.config.koi_net.first_contact.rid):
            return
        
        if self.graph.get_neighbors(
            direction="in", allowed_type=KoiNetNode):
            return
        
        self.handshake_with(self.config.koi_net.first_contact.rid)
        
    def handshake_with(self, target: KoiNetNode):
        """Initiates a handshake with target node.
        
        Pushes successive `FORGET` and `NEW` events to target node to
        reset the target's cache in case it already knew this node. 
        """
        
        log.debug(f"Initiating handshake with {target}")
        self.event_queue.push(
            Event.from_rid(
                event_type=EventType.FORGET, 
                rid=self.identity.rid),
            target=target
        )
        self.event_queue.push(
            event=Event.from_bundle(
                event_type=EventType.NEW, 
                bundle=self.cache.read(self.identity.rid)),
            target=target
        )


from enum import StrEnum


START_FUNC_NAME = "start"
STOP_FUNC_NAME = "stop"

START_ORDER_OVERRIDE = "_start_order"
STOP_ORDER_OVERRIDE = "_stop_order"

COMP_TYPE_OVERRIDE = "_comp_type"
COMP_ORDER_OVERRIDE = "_comp_order"

class CompType(StrEnum):
    SINGLETON = "SINGLETON"
    OBJECT = "OBJECT"

class CompOrder(StrEnum):
    WORKER = "WORKER"


import socket
import threading
import time
import structlog
import uvicorn
from fastapi import FastAPI, APIRouter
from fastapi.responses import JSONResponse

from koi_net.config.loader import ConfigLoader

from ..network.response_handler import ResponseHandler
from ..protocol.model_map import API_MODEL_MAP
from ..protocol.api_models import ErrorResponse
from ..protocol.errors import EXCEPTION_TO_ERROR_TYPE, ProtocolError
from ..config.full_node import FullNodeConfig

log = structlog.stdlib.get_logger()


class NodeServer(uvicorn.Server):
    """Entry point for full nodes, manages FastAPI server."""
    _config: FullNodeConfig
    response_handler: ResponseHandler
    app: FastAPI
    router: APIRouter
    
    def __init__(
        self,
        config: FullNodeConfig,
        config_loader: ConfigLoader,
        response_handler: ResponseHandler
    ):
        self._config = config
        self.config_loader = config_loader
        self.response_handler = response_handler
        
        self.build_app()
        
        self.thread = threading.Thread(target=self.run)
        
        super().__init__(config=uvicorn.Config(
            app=self.app,
            host=self._config.server.host,
            port=self._config.server.port,
            log_config=None,
            lifespan="off"
        ))
        
    def build_endpoints(self, router: APIRouter):
        """Builds endpoints for API router."""
        for path, models in API_MODEL_MAP.items():
            def create_endpoint(path: str):
                async def endpoint(req):
                    return self.response_handler.handle_response(path, req)
                
                # programmatically setting type hint annotations for FastAPI's model validation 
                endpoint.__annotations__ = {
                    "req": models.request_envelope,
                    "return": models.response_envelope
                }
                
                return endpoint
            
            router.add_api_route(
                path=path,
                endpoint=create_endpoint(path),
                methods=["POST"],
                response_model_exclude_none=True
            )
    
    def build_app(self):
        """Builds FastAPI app."""
        self.app = FastAPI(
            title="KOI-net Protocol API",
            version="1.1.0"
        )
        
        self.app.add_exception_handler(ProtocolError, self.protocol_error_handler)
        self.router = APIRouter(prefix="/koi-net")
        self.build_endpoints(self.router)
        self.app.include_router(self.router)
        
    def protocol_error_handler(self, request, exc: ProtocolError):
        """Catches `ProtocolError` and returns an `ErrorResponse` payload."""
        log.error(exc)
        resp = ErrorResponse(error=EXCEPTION_TO_ERROR_TYPE[type(exc)])
        log.info(f"Returning error response: {resp}")
        return JSONResponse(
            status_code=400,
            content=resp.model_dump(mode="json")
        )
    
    def acquire_port(self):
        derived_url = self._config.koi_net.node_profile.base_url == self._config.server.url
        
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            address = (self._config.server.host, self._config.server.port)
            while s.connect_ex(address) == 0:
                log.debug(f"port {address[1]} in use")
                self._config.server.port += 1
                address = (address[0], self._config.server.port)
        log.debug(f"acquired port {address[1]}")
        
        if derived_url:
            self._config.koi_net.node_profile.base_url = self._config.server.url
        
        self.config_loader.save_to_yaml()
    
    def start(self):
        self.acquire_port()
        self.thread.start()
        
    def stop(self):
        self.should_exit = True
        if self.thread.is_alive():
            self.thread.join()


import inspect
from collections import deque
from typing import TYPE_CHECKING, Any

import structlog

from ..exceptions import BuildError
from .consts import (
    COMP_ORDER_OVERRIDE,
    COMP_TYPE_OVERRIDE, 
    START_FUNC_NAME, 
    START_ORDER_OVERRIDE, 
    STOP_FUNC_NAME, 
    STOP_ORDER_OVERRIDE,
    CompOrder,
    CompType
)

if TYPE_CHECKING:
    from .assembler import NodeAssembler

log = structlog.stdlib.get_logger()


class BuildArtifact:
    assembler: "NodeAssembler"
    comp_dict: dict[str, Any]
    dep_graph: dict[str, list[str]]
    comp_types: dict[str, CompType]
    init_order: list[str]
    start_order: list[str]
    stop_order: list[str]
    graphviz: str
    
    def __init__(self, assembler: "NodeAssembler"):
        self.assembler = assembler
        
    def collect_comps(self):
        """Collects components from class definition."""
        
        self.comp_dict = {}
        # adds components from class and all base classes. skips `type`, and runs in reverse so that sub classes override super class values
        for base in reversed(inspect.getmro(self.assembler)[:-1]):
            for k, v in vars(base).items():
                # excludes built in, private, and `None` attributes
                if k.startswith("_") or v is None:
                    continue
                
                self.comp_dict[k] = v
        log.debug(f"Collected {len(self.comp_dict)} components")
    
    def build_dependencies(self):
        """Builds dependency graph and component type map.
        
        Graph representation is an adjacency list: the key is a component 
        name, and the value is a tuple containing names of the depedencies.
        """
        
        self.comp_types = {}
        self.dep_graph = {}
        for comp_name, comp in self.comp_dict.items():
            
            dep_names = []
            
            explicit_type = getattr(comp, COMP_TYPE_OVERRIDE, None)
            if explicit_type:
                self.comp_types[comp_name] = explicit_type
            
            # non callable components are objects treated "as is"
            elif not callable(comp):
                self.comp_types[comp_name] = CompType.OBJECT
            
            # callable components default to singletons
            else:
                sig = inspect.signature(comp)
                self.comp_types[comp_name] = CompType.SINGLETON
                dep_names = list(sig.parameters)
                
                invalid_deps = set(dep_names) - set(self.comp_dict)
                if invalid_deps:
                    raise BuildError(f"Dependencies {invalid_deps} of component '{comp_name}' are undefined")
                
            self.dep_graph[comp_name] = dep_names
        
        log.debug("Built dependency graph")
    
    def build_init_order(self):
        """Builds component initialization order using Kahn's algorithm."""
        
        # adj list: n -> outgoing neighbors
        adj = self.dep_graph
        # reverse adj list: n -> incoming neighbors
        r_adj: dict[str, list[str]] = {}
        
        # computes reverse adjacency list
        for node in adj:
            r_adj.setdefault(node, [])
            for n in adj[node]:
                r_adj.setdefault(n, [])
                r_adj[n].append(node)
        
        # how many outgoing edges each node has
        out_degree = {
            n: len(neighbors) 
            for n, neighbors in adj.items()
        }
        
        # initializing queue: nodes w/o dependencies
        queue = deque()
        for node in out_degree:
            if out_degree[node] == 0:
                queue.append(node)
        
        self.init_order = []
        while queue:
            # removes node from graph
            n = queue.popleft()
            self.init_order.append(n)
            
            # updates out degree for nodes dependent on this node
            for next_n in r_adj[n]:
                out_degree[next_n] -= 1
                # adds nodes now without dependencies to queue
                if out_degree[next_n] == 0:
                    queue.append(next_n)
        
        if len(self.init_order) != len(self.dep_graph):
            cycle_nodes = set(self.dep_graph) - set(self.init_order)
            raise BuildError(f"Found cycle in dependency graph, the following nodes could not be ordered: {cycle_nodes}")
        
        log.debug(f"Resolved initialization order: {' -> '.join(self.init_order)}")
        
    def build_start_order(self):
        """Builds component start order.
        
        Checks if components define a start function in init order. Can
        be overridden by setting start order override in the `NodeAssembler`.
        """
        
        self.start_order = getattr(self.assembler, START_ORDER_OVERRIDE, None)
        
        if self.start_order:
            return
        
        workers = []
        start_order = []
        for comp_name in self.init_order:
            comp = self.comp_dict[comp_name]
            if getattr(comp, START_FUNC_NAME, None):
                if getattr(comp, COMP_ORDER_OVERRIDE, None) == CompOrder.WORKER:
                    workers.append(comp_name)
                else:
                    start_order.append(comp_name) 
        
        # order workers first
        self.start_order = workers + start_order
        
        log.debug(f"Resolved start order: {' -> '.join(self.start_order)}")
        
    def build_stop_order(self):
        """Builds component stop order.
        
        Checks if components define a stop function in init order. Can
        be overridden by setting stop order override in the `NodeAssembler`.
        """
        self.stop_order = getattr(self.assembler, STOP_ORDER_OVERRIDE, None)
        
        if self.stop_order:
            return
        
        workers = []
        stop_order = []
        for comp_name in self.init_order:
            comp = self.comp_dict[comp_name]
            if getattr(comp, STOP_FUNC_NAME, None):
                if getattr(comp, COMP_ORDER_OVERRIDE, None) == CompOrder.WORKER:
                    workers.append(comp_name)
                else:
                    stop_order.append(comp_name) 
        
        # order workers first (last)
        self.stop_order = workers + stop_order
        # reverse order from start order
        self.stop_order.reverse()
        
        log.debug(f"Resolved stop order: {' -> '.join(self.stop_order)}")

    def visualize(self) -> str:
        """Creates representation of dependency graph in Graphviz DOT language."""
        
        s = "digraph G {\n"
        for node, neighbors in self.dep_graph.items():
            sub_s = node
            if neighbors:
                sub_s += f"-> {', '.join(neighbors)}"
            sub_s = sub_s.replace("graph", "graph_") + ";"
            s += " " * 4 + sub_s + "\n"
        s += "}"
        self.graphviz = s
    
    def build(self):
        log.debug("Creating build artifact...")
        self.collect_comps()
        self.build_dependencies()
        self.build_init_order()
        self.build_start_order()
        self.build_stop_order()
        self.visualize()
        log.debug("Done")



from typing import NamedTuple
from pydantic import BaseModel
from .envelope import SignedEnvelope
from .consts import (
    BROADCAST_EVENTS_PATH, 
    POLL_EVENTS_PATH,
    FETCH_BUNDLES_PATH, 
    FETCH_MANIFESTS_PATH, 
    FETCH_RIDS_PATH
)
from .api_models import (
    EventsPayload,
    PollEvents,
    FetchBundles,
    BundlesPayload,
    FetchManifests,
    ManifestsPayload,
    FetchRids,
    RidsPayload
)


class Models(NamedTuple):
    request: type[BaseModel]
    response: type[BaseModel] | None
    request_envelope: type[SignedEnvelope]
    response_envelope: type[SignedEnvelope] | None


# maps API paths to request and response models
API_MODEL_MAP: dict[str, Models] = {
    BROADCAST_EVENTS_PATH: Models(
        request=EventsPayload,
        response=None,
        request_envelope=SignedEnvelope[EventsPayload],
        response_envelope=None
    ),
    POLL_EVENTS_PATH: Models(
        request=PollEvents,
        response=EventsPayload,
        request_envelope=SignedEnvelope[PollEvents],
        response_envelope=SignedEnvelope[EventsPayload]
    ),
    FETCH_BUNDLES_PATH: Models(
        request=FetchBundles,
        response=BundlesPayload,
        request_envelope=SignedEnvelope[FetchBundles],
        response_envelope=SignedEnvelope[BundlesPayload]
    ),
    FETCH_MANIFESTS_PATH: Models(
        request=FetchManifests,
        response=ManifestsPayload,
        request_envelope=SignedEnvelope[FetchManifests],
        response_envelope=SignedEnvelope[ManifestsPayload]
    ),
    FETCH_RIDS_PATH: Models(
        request=FetchRids,
        response=RidsPayload,
        request_envelope=SignedEnvelope[FetchRids],
        response_envelope=SignedEnvelope[RidsPayload]
    )
}



import threading
import time
import structlog

from ..processor.kobj_queue import KobjQueue
from ..network.resolver import NetworkResolver
from ..config.partial_node import PartialNodeConfig

log = structlog.stdlib.get_logger()


class NodePoller:
    """Entry point for partial nodes, manages polling event loop."""
    kobj_queue: KobjQueue
    resolver: NetworkResolver
    config: PartialNodeConfig
    
    def __init__(
        self,
        config: PartialNodeConfig,
        kobj_queue: KobjQueue,
        resolver: NetworkResolver
    ):
        self.kobj_queue = kobj_queue
        self.resolver = resolver
        self.config = config
        self.exit_event = threading.Event()
        
        self.thread = threading.Thread(target=self.run)

    def poll(self):
        """Polls neighbor nodes and processes returned events."""
        for node_rid, events in self.resolver.poll_neighbors().items():
            for event in events:
                self.kobj_queue.push(event=event, source=node_rid)

    def run(self):
        """Runs polling event loop."""
        while not self.exit_event.is_set():
            start_time = time.monotonic()
            self.poll()
            elapsed = time.monotonic() - start_time
            wait_time = max(0, self.config.poller.polling_interval - elapsed)
            self.exit_event.wait(wait_time)
                
    def start(self):
        self.thread.start()
        
    def stop(self):
        self.exit_event.set()
        if self.thread.is_alive():
            self.thread.join()


class KoiNetCliError(Exception):
    ...

class MissingEnvVariablesError(KoiNetCliError):
    def __init__(self, message: str, vars: list[str]):
        super().__init__(message)
        self.vars = vars
        
class NodeExistsError(KoiNetCliError):
    ...
    
class NodeNotFoundError(KoiNetCliError):
    ...


from pydantic import BaseModel
from rid_lib import RIDType
from rid_lib.types import KoiNetNode

from ..protocol.node import NodeProfile


class EventWorkerConfig(BaseModel):
    queue_timeout: float = 0.1
    max_buf_len: int = 5
    max_wait_time: float = 1.0
    
class KobjWorkerConfig(BaseModel):
    queue_timeout: float = 0.1

class NodeContact(BaseModel):
    rid: KoiNetNode | None = None
    url: str | None = None

class KoiNetConfig(BaseModel):
    """Config for KOI-net parameters."""
    
    node_name: str
    node_rid: KoiNetNode | None = None
    node_profile: NodeProfile
    
    rid_types_of_interest: list[RIDType] = [KoiNetNode]
        
    cache_directory_path: str = ".rid_cache"
    private_key_pem_path: str = "priv_key.pem"
    
    event_worker: EventWorkerConfig = EventWorkerConfig()
    kobj_worker: KobjWorkerConfig = KobjWorkerConfig()
    
    first_contact: NodeContact = NodeContact()


from rid_lib.types import KoiNetNode
from .config.base import BaseNodeConfig
from .protocol.node import NodeProfile


class NodeIdentity:
    """Represents a node's identity (RID, profile)."""
    
    config: BaseNodeConfig
    
    def __init__(self, config: BaseNodeConfig):
        self.config = config
        
    @property
    def rid(self) -> KoiNetNode:
        return self.config.koi_net.node_rid
    
    @property
    def profile(self) -> NodeProfile:
        return self.config.koi_net.node_profile



import time
import typer
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from dotenv import load_dotenv

from ..interfaces.network import NetworkInterface


load_dotenv()


app = typer.Typer()
console = Console()

@app.command()
def list_node_types():
    network = NetworkInterface()
    
    table = Table()
    table.add_column("aliases", style="cyan")
    table.add_column("node types", style="magenta")

    for module, aliases in network.get_node_modules().items():
        table.add_row(", ".join(aliases), module)
    console.print(table)
    
@app.command()
def list_nodes():
    table = Table(title="created nodes")
    table.add_column("name", style="cyan")
    table.add_column("type", style="green")
    table.add_column("rid", style="magenta")

    network = NetworkInterface()
    for name, node in network.nodes.items():
        node_conf = node.get_config()
        table.add_row(name, node.module, str(node_conf.koi_net.node_rid))
        
    console.print(table)
    
@app.command()
def start(delay: int = 1):
    network = NetworkInterface()
    print("starting network...")
    network.start(delay=delay)
    
    try:
        while any(n.process.poll() is None for n in network.nodes.values()):
            time.sleep(0.1)
    except KeyboardInterrupt:
        print("stopping network...")
        network.stop()
        
        for node in network.nodes.values():
            node.process.wait()
    
@app.command()
def set_first_contact(name: str, force: bool = False):
    network = NetworkInterface()

    print(f"First contact updated from '{network.config.first_contact}' -> '{name}'")
    
    network.config.first_contact = name
    network.config_loader.save_to_yaml()
    
    fc_node = network.nodes[network.config.first_contact]
    fc_config = fc_node.get_config()
    fc_rid = fc_config.koi_net.node_rid
    fc_url = fc_config.koi_net.node_profile.base_url
    
    updated_nodes = 0
    for node in network.nodes.values():
        with node.mutate_config() as n_config:
            if not force and n_config.koi_net.first_contact.rid:
                continue
            
            if node.name == network.config.first_contact:
                continue
            
            n_config.koi_net.first_contact.rid = fc_rid
            n_config.koi_net.first_contact.url = fc_url
            updated_nodes += 1
    
    print(f"Updated config for {updated_nodes} node(s)")
        


import threading
import structlog

from .artifact import BuildArtifact
from .consts import START_FUNC_NAME, STOP_FUNC_NAME

log = structlog.stdlib.get_logger()


class NodeContainer:
    """Dummy 'shape' for node containers built by assembler."""
    _artifact: BuildArtifact
    
    shutdown_event: threading.Event
    
    def __init__(self, artifact, **kwargs):
        self._artifact = artifact
        
        # adds all components as attributes of this instance
        for name, comp in kwargs.items():
            setattr(self, name, comp)
    
    def run(self):
        try:
            self.start()
            self.shutdown_event.wait()
        except KeyboardInterrupt:
            log.info("Received keyboard interrupt")
            self.shutdown_event.set()
        finally:
            self.stop()
    
    def start(self):
        log.info("Starting node...")
        for comp_name in self._artifact.start_order:
            comp = getattr(self, comp_name)
            start_func = getattr(comp, START_FUNC_NAME)
            log.info(f"Starting {comp_name}...")
            start_func()
            
    def stop(self):
        log.info("Stopping node...")
        for comp_name in self._artifact.stop_order:
            comp = getattr(self, comp_name)
            stop_func = getattr(comp, STOP_FUNC_NAME)
            log.info(f"Stopping {comp_name}...")
            stop_func()



import pkgutil
import importlib
from importlib.metadata import entry_points
import time

from koi_net.config.proxy import ConfigProxy
from koi_net.exceptions import NodeNotFoundError

from ..models import NetworkConfigLoader, KoiNetworkConfig
from .node import NodeInterface

ENTRY_POINT_GROUP = "koi_net.node"
MODULE_PREFIX = "koi_net_"
MODULE_POSTFIX = "_node"


class NetworkInterface:
    def __init__(self):
        self.config: KoiNetworkConfig = ConfigProxy()
        self.config_loader = NetworkConfigLoader(
            config_schema=KoiNetworkConfig,
            config=self.config
        )
        
        self.nodes: dict[str, NodeInterface] = {}
        
        self.load_nodes()
    
    def load_nodes(self):
        for name, module in self.config.nodes.items():
            self.nodes[name] = NodeInterface(name, module)
            
    def start(self, delay: float = 0.0):
        for name, node in self.nodes.items():
            print(f"starting {name}...")
            node.start()
            time.sleep(delay)
            
    def stop(self):
        for name, node in self.nodes.items():
            print(f"stopping {name}...")
            node.stop()
        
    def qualify_node_reference(self, node_module_ref: str) -> str | None:
        eps = entry_points(group=ENTRY_POINT_GROUP, name=node_module_ref)
        if len(eps) == 0:
            try:
                importlib.import_module(node_module_ref)
                return node_module_ref
            except ImportError:
                try:
                    expanded_ref = MODULE_PREFIX + node_module_ref + MODULE_POSTFIX
                    importlib.import_module(expanded_ref)
                    return expanded_ref
                except ImportError:
                    raise Exception(f"no node module of name '{node_module_ref}' exists")
            
        elif len(eps) == 1:
            ep, = eps
            return ep.module
        
        else:
            raise Exception("more than one endpoint with that name found")
        
    def get_node_modules(self) -> dict[str, set[str]]:
        module_map = {
            ep.module: {ep.name} for ep in entry_points(group=ENTRY_POINT_GROUP)
        }
        for module in pkgutil.iter_modules():
            if not (module.name.startswith(MODULE_PREFIX) 
                and module.name.endswith(MODULE_POSTFIX)):
                continue
            
            module_alias = module.name[len(MODULE_PREFIX):-len(MODULE_POSTFIX)]
            module_map.setdefault(module.name, set()).add(module_alias)
            
        
        return module_map
        
    def create_node(self, node_name: str, node_module_ref: str | None = None):
        node_module = self.qualify_node_reference(node_module_ref or node_name)
        
        self.nodes[node_name] = NodeInterface.create(node_name, node_module)
        
        self.config.nodes[node_name] = node_module
        self.config_loader.save_to_yaml()
        
    def delete_node(self, node_name: str):
        if node_name not in self.nodes:
            raise NodeNotFoundError(f"Node '{node_name}' not found")
        
        self.nodes[node_name].delete()
        del self.nodes[node_name]
        del self.config.nodes[node_name]
        self.config_loader.save_to_yaml()
        
        
if __name__ == "__main__":
    network = NetworkInterface()


import threading

from ..build import comp_order


class End:
    """Class for STOP_WORKER sentinel pushed to worker queues."""
    pass

STOP_WORKER = End()

@comp_order.worker
class ThreadWorker:
    """Base class for thread workers."""
    
    thread: threading.Thread
    
    def __init__(self):
        self.thread = threading.Thread(target=self.run)
        
    def start(self):
        self.thread.start()
        
    def stop(self):
        if self.thread.is_alive():
            self.thread.join()
        
    def run(self):
        """Processing loop for thread."""
        pass


from typing import Generic, TypeVar


T = TypeVar("T")

class ConfigProxy(Generic[T]):
    """Proxy for config access.
    
    Allows initialization of this component, and updating state without
    destroying the original reference. Handled as if it were a config
    model by other classes, loaded and saved by the `ConfigLoader`.
    """
    _delegate: T
    
    def __init__(self):
        object.__setattr__(self, "_delegate", None)
    
    def _set_delegate(self, delegate: T):
        object.__setattr__(self, "_delegate", delegate)
    
    def _get_delegate(self) -> T:
        delegate = object.__getattribute__(self, "_delegate")
        if delegate is None:
            raise RuntimeError("Proxy called before delegate loaded")
        return delegate
    
    def __getattr__(self, name):
        return getattr(self._get_delegate(), name)
    
    def __setattr__(self, name, value):
        delegate = self._get_delegate()
        setattr(delegate, name, value)



import structlog
from rid_lib import RID
from rid_lib.types import KoiNetNode
from rid_lib.ext import Manifest, Cache
from rid_lib.ext.bundle import Bundle

from ..processor.kobj_queue import KobjQueue
from ..protocol.consts import BROADCAST_EVENTS_PATH, FETCH_BUNDLES_PATH, FETCH_MANIFESTS_PATH, FETCH_RIDS_PATH, POLL_EVENTS_PATH
from ..protocol.envelope import SignedEnvelope
from ..secure_manager import SecureManager
from .event_buffer import EventBuffer


from ..protocol.api_models import (
    EventsPayload,
    PollEvents,
    RidsPayload,
    ManifestsPayload,
    BundlesPayload,
    FetchRids,
    FetchManifests,
    FetchBundles,
)

log = structlog.stdlib.get_logger()


class ResponseHandler:
    """Handles generating responses to requests from other KOI nodes."""
    
    cache: Cache
    kobj_queue: KobjQueue
    poll_event_buf: EventBuffer
    
    def __init__(
        self, 
        cache: Cache,
        kobj_queue: KobjQueue,
        poll_event_buf: EventBuffer,
        secure_manager: SecureManager
    ):
        self.cache = cache
        self.kobj_queue = kobj_queue
        self.poll_event_buf = poll_event_buf
        self.secure_manager = secure_manager
    
    def handle_response(self, path: str, req: SignedEnvelope):
        self.secure_manager.validate_envelope(req)
        
        response_map = {
            BROADCAST_EVENTS_PATH: self.broadcast_events_handler,
            POLL_EVENTS_PATH: self.poll_events_handler,
            FETCH_RIDS_PATH: self.fetch_rids_handler,
            FETCH_MANIFESTS_PATH: self.fetch_manifests_handler,
            FETCH_BUNDLES_PATH: self.fetch_bundles_handler
        }
        
        response = response_map[path](req.payload, req.source_node)
        
        if response is None:
            return
        
        return self.secure_manager.create_envelope(
            payload=response,
            target=req.source_node
        )
        
    def broadcast_events_handler(self, req: EventsPayload, source: KoiNetNode):
        log.info(f"Request to broadcast events, received {len(req.events)} event(s)")
        
        for event in req.events:
            self.kobj_queue.push(event=event, source=source)
        
    def poll_events_handler(
        self, 
        req: PollEvents, 
        source: KoiNetNode
    ) -> EventsPayload:
        events = self.poll_event_buf.flush(source, limit=req.limit)
        log.info(f"Request to poll events, returning {len(events)} event(s)")
        return EventsPayload(events=events)
        
    def fetch_rids_handler(
        self, 
        req: FetchRids, 
        source: KoiNetNode
    ) -> RidsPayload:
        """Returns response to fetch RIDs request."""
        rids = self.cache.list_rids(req.rid_types)
        log.info(f"Request to fetch rids, allowed types {req.rid_types}, returning {len(rids)} RID(s)")
        return RidsPayload(rids=rids)
        
    def fetch_manifests_handler(
        self, 
        req: FetchManifests, 
        source: KoiNetNode
    ) -> ManifestsPayload:
        """Returns response to fetch manifests request."""        
        manifests: list[Manifest] = []
        not_found: list[RID] = []
        
        for rid in (req.rids or self.cache.list_rids(req.rid_types)):
            bundle = self.cache.read(rid)
            if bundle:
                manifests.append(bundle.manifest)
            else:
                not_found.append(rid)
        
        log.info(f"Request to fetch manifests, allowed types {req.rid_types}, rids {req.rids}, returning {len(manifests)} manifest(s)")
        return ManifestsPayload(manifests=manifests, not_found=not_found)
        
    def fetch_bundles_handler(
        self, 
        req: FetchBundles, 
        source: KoiNetNode
    ) -> BundlesPayload:
        """Returns response to fetch bundles request."""
        
        bundles: list[Bundle] = []
        not_found: list[RID] = []

        for rid in req.rids:
            bundle = self.cache.read(rid)
            if bundle:
                bundles.append(bundle)
            else:
                not_found.append(rid)
                
        log.info(f"Request to fetch bundles, requested rids {req.rids}, returning {len(bundles)} bundle(s)")
        return BundlesPayload(bundles=bundles, not_found=not_found)


"""KOI-net library exceptions.

Exception hierarchy map:
- `KoiNetError`
  - `BuildError`
  - `RequestError`
    - `ClientError`
      - `SelfRequestError`
      - `PartialNodeQueryError`
      - `NodeNotFoundError`
    - `TransportError`
    - `ServerError`
      - `RemoteProtocolError`
        - `RemoteUnknownNodeError`
        - `RemoteInvalidKeyError`
        - `RemoteInvalidSignatureError`
        - `RemoteInvalidTargetError`
  - `ProtocolError`
    - `UnknownNodeError`
    - `InvalidKeyError`
    - `InvalidSignatureError`
    - `InvalidTargetError`
"""


# BASE EXCEPTION
class KoiNetError(Exception):
    """Base exception."""
    pass

# BUILD ERRORS
class BuildError(KoiNetError):
    """Raised when errors occur in build process."""
    pass

# NETWORK REQUEST ERRORS
class RequestError(KoiNetError):
    """Base for network request errors."""
    pass

# CLIENT ERRORS
class ClientError(RequestError):
    """Raised when this node makes an invalid request."""
    pass

class SelfRequestError(ClientError):
    """Raised when this node tries to request itself."""
    pass

class PartialNodeQueryError(ClientError):
    """Raised when this node attempts to query a partial node."""
    pass

class NodeNotFoundError(ClientError):
    """Raised when this node cannot find a node's URL."""
    pass

class TransportError(RequestError):
    """Raised when a transport error occurs during a request."""
    pass

# SERVER ERRORS
class ServerError(RequestError):
    """Raised when an server error occurs during a request."""
    pass

# PROTOCOL ERRORS
class RemoteProtocolError(ServerError):
    """Base for protocol errors raised by peer node."""
    pass

class RemoteUnknownNodeError(RemoteProtocolError):
    """Raised by peer node when this node is unknown."""
    pass
    
class RemoteInvalidKeyError(RemoteProtocolError):
    """Raised by peer node when this node's public key doesn't match their RID."""
    pass
    
class RemoteInvalidSignatureError(RemoteProtocolError):
    """Raised by peer node when this node's envelope signature is invalid."""
    pass

class RemoteInvalidTargetError(RemoteProtocolError):
    """Raised by peer node when this node's envelope target is not it's RID."""
    pass


class ProtocolError(KoiNetError):
    """Base for protocol errors raised by this node."""
    pass

class UnknownNodeError(ProtocolError):
    """Raised when peer node is unknown."""
    pass
    
class InvalidKeyError(ProtocolError):
    """Raised when peer node's public key doesn't match their RID."""
    pass
    
class InvalidSignatureError(ProtocolError):
    """Raised when peer node's envelope signature is invalid."""
    pass

class InvalidTargetError(ProtocolError):
    """Raised when peer node's target is not this node."""
    pass


import time
from contextlib import contextmanager
from typing import Generator
from rid_lib.types import KoiNetNode

from ..protocol.event import Event


class EventBuffer:
    """Stores outgoing events sent to other nodes."""

    buffers: dict[KoiNetNode, list[Event]]
    start_time: dict[KoiNetNode, float]
    
    def __init__(self):
        self.buffers = {}
        self.start_time = {}
        
    def push(self, node: KoiNetNode, event: Event):
        """Pushes event to specified node.
        
        Sets start time to now if unset.
        """
        
        self.start_time.setdefault(node, time.time())
        
        event_buf = self.buffers.setdefault(node, [])
        event_buf.append(event)
        
    def buf_len(self, node: KoiNetNode):
        """Returns the length of a node's event buffer."""
        return len(self.buffers.get(node, []))
        
    def flush(self, node: KoiNetNode, limit: int = 0) -> list[Event]:
        """Flushes all (or limit) events for a node.
        
        Resets start time.
        """
        self.start_time.pop(node, None)
        
        if node not in self.buffers:
            return []
        
        event_buf = self.buffers[node]
        
        if limit and len(event_buf) > limit:
            flushed_events = event_buf[:limit]
            self.buffers[node] = event_buf[limit:]
        else:
            flushed_events = event_buf.copy()
            del self.buffers[node]
        
        return flushed_events
    
    @contextmanager
    def safe_flush(
        self, 
        node: KoiNetNode, 
        limit: int = 0,
        force_flush: bool = False
    ) -> Generator[list[Event], None, None]:
        """Context managed safe flush, only commits on successful exit.
        
        Exceptions will result in buffer rollback to the previous state.
        """
        
        self.start_time.pop(node, None)
        
        if node not in self.buffers:
            yield []
            return
        
        event_buf = self.buffers[node].copy()
        in_place = limit and len(event_buf) > limit
        
        try:
            if in_place:
                yield event_buf[:limit]
                self.buffers[node] = event_buf[limit:]
            else:
                yield event_buf.copy()
                self.buffers.pop(node, None)
        
        except Exception:
            # if force, flushes buffers and reraises exception
            if force_flush:
                if in_place:
                    self.buffers[node] = event_buf[limit:]
                else:
                    self.buffers.pop(node, None)
            raise



from pydantic import BaseModel, model_validator
from .base import BaseNodeConfig, KoiNetConfig as BaseKoiNetConfig
from ..protocol.node import (
    NodeProfile as BaseNodeProfile, 
    NodeType, 
    NodeProvides
)


class NodeProfile(BaseNodeProfile):
    """Node profile config class for full nodes."""
    node_type: NodeType = NodeType.FULL

class KoiNetConfig(BaseKoiNetConfig):
    """KOI-net config class for full nodes."""
    node_profile: NodeProfile

class ServerConfig(BaseModel):
    """Server config for full nodes.
    
    The parameters in this class represent how a server should be hosted,
    not accessed. For example, a node may host a server at
    `http://127.0.0.1:8000/koi-net`, but serve through nginx at
    `https://example.com/koi-net`.
    """
    
    host: str = "127.0.0.1"
    port: int = 8000
    path: str | None = "/koi-net"
    
    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}{self.path or ''}"

class FullNodeConfig(BaseNodeConfig):
    """Node config class for full nodes."""
    koi_net: KoiNetConfig
    server: ServerConfig = ServerConfig()
    
    @model_validator(mode="after")
    def check_url(self):
        """Generates base URL if missing from node profile."""
        if not self.koi_net.node_profile.base_url:
            self.koi_net.node_profile.base_url = self.server.url
        return self



from pydantic import BaseModel
from rid_lib import RID
from rid_lib.ext import Manifest, Bundle
from rid_lib.types import KoiNetNode
from ..protocol.event import Event, EventType


class KnowledgeObject(BaseModel):
    """A normalized knowledge representation for internal processing.
    
    Represents an RID, manifest, bundle, or event. Contains three fields 
    (`normalized_event_type`, `source`, `network_targets`) used for 
    decision making in the knowledge processing pipeline. 
    
    The source indicates which node this object originated from, or
    `None` if it was generated by this node.
    
    The normalized event type indicates how the knowledge object is 
    viewed from the perspective of this node, and what cache actions 
    will take place. (`NEW`, `UPDATE`) -> cache write, `FORGET` -> 
    cache delete, `None` -> no cache action, and end of pipeline.
    
    The network targets indicate other nodes in the network this 
    knowledge object will be sent to. The event sent to them will be 
    constructed from this knowledge object's RID, manifest, contents, 
    and normalized event type.
    """
    
    rid: RID
    manifest: Manifest | None = None
    contents: dict | None = None
    event_type: EventType | None = None
    normalized_event_type: EventType | None = None
    source: KoiNetNode | None = None
    network_targets: set[KoiNetNode] = set()
    
    def __repr__(self):
        return f"<KObj '{self.rid}' event type: '{self.event_type}' -> '{self.normalized_event_type}', source: '{self.source}'>"
    
    @classmethod
    def from_rid(
        cls, 
        rid: RID, 
        event_type: EventType | None = None, 
        source: KoiNetNode | None = None
    ) -> "KnowledgeObject":
        """Creates a `KnowledgeObject` from an `RID`."""
        
        return cls(
            rid=rid,
            event_type=event_type,
            source=source
        )
        
    @classmethod
    def from_manifest(
        cls, 
        manifest: Manifest, 
        event_type: EventType | None = None, 
        source: KoiNetNode | None = None
    ) -> "KnowledgeObject":
        """Creates a `KnowledgeObject` from a `Manifest`."""
        
        return cls(
            rid=manifest.rid,
            manifest=manifest,
            event_type=event_type,
            source=source
        )
        
    @classmethod
    def from_bundle(
        cls, 
        bundle: Bundle, 
        event_type: EventType | None = None, 
        source: KoiNetNode | None = None
    ) -> "KnowledgeObject":
        """Creates a `KnowledgeObject` from a `Bundle`."""
        
        return cls(
            rid=bundle.rid,
            manifest=bundle.manifest,
            contents=bundle.contents,
            event_type=event_type,
            source=source
        )
        
    @classmethod
    def from_event(
        cls,
        event: Event,
        source: KoiNetNode | None = None
    ) -> "KnowledgeObject":
        """Creates a `KnowledgeObject` from an `Event`."""
        
        return cls(
            rid=event.rid,
            manifest=event.manifest,
            contents=event.contents,
            event_type=event.event_type,
            source=source
        )
    
    @property
    def bundle(self) -> Bundle:
        """Bundle representation of knowledge object."""
        
        if self.manifest is None or self.contents is None:
            raise ValueError("Knowledge object missing manifest or contents, cannot convert to `Bundle`.")
        
        return Bundle(
            manifest=self.manifest,
            contents=self.contents
        )
    
    @property
    def normalized_event(self) -> Event:
        """Event representation of knowledge object."""
        
        if self.normalized_event_type is None:
            raise ValueError("Internal event's normalized event type is None, cannot convert to Event")
        
        elif self.normalized_event_type == EventType.FORGET:
            return Event(
                rid=self.rid,
                event_type=EventType.FORGET
            )
        
        else:
            return Event(
                rid=self.rid,
                event_type=self.normalized_event_type,
                manifest=self.manifest,
                contents=self.contents
            )


import structlog
from pydantic import BaseModel, Field, model_validator

from ..build import comp_type
from ..protocol.secure import PrivateKey
from .env_config import EnvConfig
from .koi_net_config import KoiNetConfig

log = structlog.stdlib.get_logger()


@comp_type.object
class BaseNodeConfig(BaseModel):
    """Base node config class, intended to be extended.
    
    Using the `comp_type.object` decorator to mark this class as an
    object to be treated "as is" rather than attempting to initialize it
    during the build.
    """
    
    koi_net: KoiNetConfig
    # note: EnvConfig has to use a default factory, otherwise it will
    # evaluated during the library import and cause an error if any
    # env variables are undefined
    env: EnvConfig = Field(default_factory=EnvConfig)
    
    @model_validator(mode="after")
    def generate_rid_cascade(self):
        """Generates node RID if missing."""
        if self.koi_net.node_rid and self.koi_net.node_profile.public_key:
            return self
        
        log.debug("Node RID or public key not found in config, attempting to generate")
        
        try:
            # attempts to read existing private key PEM file
            with open(self.koi_net.private_key_pem_path, "r") as f:
                priv_key_pem = f.read()
                priv_key = PrivateKey.from_pem(
                    priv_key_pem,
                    password=self.env.priv_key_password)
                log.debug("Used existing private key from PEM file")
        
        except FileNotFoundError:
            # generates new private key if PEM not found
            priv_key = PrivateKey.generate()
            
            with open(self.koi_net.private_key_pem_path, "w") as f:
                f.write(priv_key.to_pem(self.env.priv_key_password))
            log.debug("Generated new private key, no PEM file found")
        
        pub_key = priv_key.public_key()
        self.koi_net.node_rid = pub_key.to_node_rid(self.koi_net.node_name)
        log.debug(f"Node RID set to {self.koi_net.node_rid}")
        
        if self.koi_net.node_profile.public_key != pub_key.to_der():
            if self.koi_net.node_profile.public_key:
                log.warning("New private key overwriting old public key!")
            
            self.koi_net.node_profile.public_key = pub_key.to_der()
        
        return self


import sys
import threading
from koi_net.build.assembler import NodeAssembler


class ControlLoop:
    shutdown_event: threading.Event
    
    def __init__(self, shutdown_event):
        self.shutdown_event = shutdown_event
        self.thread = threading.Thread(target=self.run, daemon=True)
        
    def run(self):
        for line in sys.stdin:
            if line.strip() == "STOP":
                self.shutdown_event.set()
                
        self.shutdown_event.set()
        
    def start(self):
        self.thread.start()

class BaseAssembly(NodeAssembler):
    shutdown_event: threading.Event = threading.Event
    control_loop: ControlLoop = ControlLoop



from typing import Any, Self

import structlog

from ..exceptions import BuildError
from .artifact import BuildArtifact, CompType
from .container import NodeContainer

log = structlog.stdlib.get_logger()


class NodeAssembler:
    _artifact: BuildArtifact = None
    
    # optional order overrides:
    _start_order: list[str]
    _stop_order: list[str]
    
    # annotation hack to show the components and container methods
    def __new__(cls) -> Self | NodeContainer:
        """Returns assembled node container."""
        
        log.debug(f"Assembling '{cls.__name__}'")
        
        # builds assembly artifact if it doesn't exist
        if not cls._artifact:
            cls._artifact = BuildArtifact(cls)
            cls._artifact.build()
        
        components = cls._build_components(cls._artifact)
        
        log.debug("Returning assembled node")
        return NodeContainer(cls._artifact, **components)
    
    @staticmethod
    def _build_components(artifact: BuildArtifact):
        """Returns assembled components as a dict."""
        
        log.debug("Building components...")
        components: dict[str, Any] = {}
        for comp_name in artifact.init_order:
        # for comp_name, (comp_type, dep_names) in dep_graph.items():
            comp = artifact.comp_dict[comp_name]
            comp_type = artifact.comp_types[comp_name]
            
            if comp_type == CompType.OBJECT:
                components[comp_name] = comp
            
            elif comp_type == CompType.SINGLETON:
                # builds depedency dict for current component
                dependencies = {}
                for dep in artifact.dep_graph[comp_name]:
                    if dep not in components:
                        raise BuildError(f"Couldn't find required component '{dep}'")
                    dependencies[dep] = components[dep]
                components[comp_name] = comp(**dependencies)
        log.debug("Done")
        
        return components



from functools import wraps

import structlog
import httpx
from rid_lib import RID
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode
from pydantic import ValidationError

from ..identity import NodeIdentity
from ..protocol.api_models import (
    RidsPayload,
    ManifestsPayload,
    BundlesPayload,
    EventsPayload,
    FetchRids,
    FetchManifests,
    FetchBundles,
    PollEvents,
    RequestModels,
    ResponseModels,
    ErrorResponse
)
from ..protocol.consts import (
    BROADCAST_EVENTS_PATH,
    POLL_EVENTS_PATH,
    FETCH_RIDS_PATH,
    FETCH_MANIFESTS_PATH,
    FETCH_BUNDLES_PATH
)
from ..protocol.errors import ErrorType
from ..protocol.node import NodeProfile, NodeType
from ..protocol.model_map import API_MODEL_MAP
from ..secure_manager import SecureManager
from ..exceptions import (
    RemoteInvalidKeyError,
    RemoteInvalidSignatureError,
    RemoteInvalidTargetError,
    RequestError,
    SelfRequestError,
    PartialNodeQueryError,
    NodeNotFoundError,
    ServerError,
    TransportError,
    RemoteUnknownNodeError
)
from .error_handler import ErrorHandler

log = structlog.stdlib.get_logger()


def report_exception(func):
    """Logs request errors as warnings."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestError as err:
            log.warning(err)
            raise
    return wrapper

class RequestHandler:
    """Handles making requests to other KOI nodes."""
    
    cache: Cache
    identity: NodeIdentity
    secure_manager: SecureManager
    error_handler: ErrorHandler
    
    def __init__(
        self, 
        cache: Cache,
        identity: NodeIdentity,
        secure_manager: SecureManager,
        error_handler: ErrorHandler
    ):
        self.cache = cache
        self.identity = identity
        self.secure_manager = secure_manager
        self.error_handler = error_handler
    
    def get_base_url(self, node_rid: KoiNetNode) -> str:
        """Retrieves URL of a node from its RID."""
        
        node_bundle = self.cache.read(node_rid)
        if node_bundle:
            node_profile = node_bundle.validate_contents(NodeProfile)
            if node_profile.node_type != NodeType.FULL:
                raise PartialNodeQueryError("Partial nodes don't have URLs")
            node_url = node_profile.base_url
        
        elif node_rid == self.identity.config.koi_net.first_contact.rid:
            node_url = self.identity.config.koi_net.first_contact.url
        
        else:
            raise NodeNotFoundError(f"URL not found for {node_rid!r}")
        
        log.debug(f"Resolved {node_rid!r} to {node_url}")
        return node_url
    
    @report_exception
    def make_request(
        self,
        node: KoiNetNode,
        path: str, 
        request: RequestModels,
    ) -> ResponseModels | None:
        """Makes a request to a node."""
        if node == self.identity.rid:
            raise SelfRequestError("Don't talk to yourself")
        
        url = self.get_base_url(node) + path
        log.info(f"Making request to {url}")
    
        signed_envelope = self.secure_manager.create_envelope(
            payload=request,
            target=node
        )
        
        data = signed_envelope.model_dump_json(exclude_none=True)
        
        try:
            result = httpx.post(url, data=data)
            result.raise_for_status()
            self.error_handler.reset_timeout_counter(node)
            
        except httpx.RequestError as e:
            log.debug("Failed to connect")
            self.error_handler.handle_connection_error(node)
            raise TransportError(e)
        
        except httpx.HTTPStatusError:
            """Possible errors:
            
            4xx - KOI-net protocol error, validate body
            404/405 - not implementing endpoints, or misconfigured URL
            
            500 - internal server error
            """
            try:
                resp = ErrorResponse.model_validate_json(result.text)
                self.error_handler.handle_protocol_error(resp.error, node)
                
                match resp.error:
                    case ErrorType.UnknownNode:
                        raise RemoteUnknownNodeError(f"Peer couldn't resolve this node's RID")
                    case ErrorType.InvalidKey:
                        raise RemoteInvalidKeyError(f"Peer marked this node's public key as invalid")
                    case ErrorType.InvalidSignature:
                        raise RemoteInvalidSignatureError("Peer marked envelope signature as invalid")
                    case ErrorType.InvalidTarget:
                        raise RemoteInvalidTargetError("Envelope target is not the peer node")
            
            except ValidationError as e:
                raise ServerError(e)
            
        
        resp_env_model = API_MODEL_MAP[path].response_envelope
        if not resp_env_model:
            return
        
        try:
            resp_envelope = resp_env_model.model_validate_json(result.text)
        except ValidationError as e:
            raise ServerError(e)
        
        self.secure_manager.validate_envelope(resp_envelope)
        
        return resp_envelope.payload
    
    def broadcast_events(
        self, 
        node: RID, 
        req: EventsPayload | None = None,
        **kwargs
    ) -> None:
        """Broadcasts events to a node.
        
        Pass `EventsPayload` object, or see `protocol.api_models.EventsPayload` for available kwargs.
        """
        request = req or EventsPayload.model_validate(kwargs)
        self.make_request(node, BROADCAST_EVENTS_PATH, request)
        log.info(f"Broadcasted {len(request.events)} event(s) to {node!r}")
        
    def poll_events(
        self, 
        node: RID, 
        req: PollEvents | None = None,
        **kwargs
    ) -> EventsPayload:
        """Polls events from a node.
        
        Pass `PollEvents` object as `req` or fields as kwargs.
        """
        request = req or PollEvents.model_validate(kwargs)
        resp = self.make_request(node, POLL_EVENTS_PATH, request)
        log.info(f"Polled {len(resp.events)} events from {node!r}")
        return resp
        
    def fetch_rids(
        self, 
        node: RID, 
        req: FetchRids | None = None,
        **kwargs
    ) -> RidsPayload:
        """Fetches RIDs from a node.
        
        Pass `FetchRids` object as `req` or fields as kwargs.
        """
        request = req or FetchRids.model_validate(kwargs)
        resp = self.make_request(node, FETCH_RIDS_PATH, request)
        log.info(f"Fetched {len(resp.rids)} RID(s) from {node!r}")
        return resp
                
    def fetch_manifests(
        self, 
        node: RID, 
        req: FetchManifests | None = None,
        **kwargs
    ) -> ManifestsPayload:
        """Fetches manifests from a node.
        
        Pass `FetchManifests` object as `req` or fields as kwargs.
        """
        request = req or FetchManifests.model_validate(kwargs)
        resp = self.make_request(node, FETCH_MANIFESTS_PATH, request)
        log.info(f"Fetched {len(resp.manifests)} manifest(s) from {node!r}")
        return resp
                
    def fetch_bundles(
        self, 
        node: RID, 
        req: FetchBundles | None = None,
        **kwargs
    ) -> BundlesPayload:
        """Fetches bundles from a node.
        
        Pass `FetchBundles` object as `req` or fields as kwargs.
        """
        request = req or FetchBundles.model_validate(kwargs)
        resp = self.make_request(node, FETCH_BUNDLES_PATH, request)
        log.info(f"Fetched {len(resp.bundles)} bundle(s) from {node!r}")
        return resp


import structlog
from rid_lib import RID
from rid_lib.core import RIDType
from rid_lib.ext import Cache, Bundle
from rid_lib.types import KoiNetNode

from .graph import NetworkGraph
from .request_handler import RequestHandler
from ..protocol.node import NodeProfile, NodeType
from ..protocol.event import Event
from ..identity import NodeIdentity
from ..config.base import BaseNodeConfig
from ..exceptions import RequestError

log = structlog.stdlib.get_logger()


class NetworkResolver:
    """Handles resolving nodes or knowledge objects from the network."""
    
    config: BaseNodeConfig    
    identity: NodeIdentity
    cache: Cache
    graph: NetworkGraph
    request_handler: RequestHandler
    
    def __init__(
        self, 
        config: BaseNodeConfig,
        cache: Cache, 
        identity: NodeIdentity,
        graph: NetworkGraph,
        request_handler: RequestHandler,
    ):
        self.config = config
        self.identity = identity
        self.cache = cache
        self.graph = graph
        self.request_handler = request_handler
        
        self.poll_event_queue = dict()
        self.webhook_event_queue = dict()
    
    def get_state_providers(self, rid_type: RIDType) -> list[KoiNetNode]:
        """Returns list of node RIDs which provide state for specified RID type."""
        
        log.debug(f"Looking for state providers of {rid_type}")
        provider_nodes = []
        for node_rid in self.cache.list_rids(rid_types=[KoiNetNode]):
            if node_rid == self.identity.rid:
                continue
            
            node_bundle = self.cache.read(node_rid)
            node_profile = node_bundle.validate_contents(NodeProfile)
            
            if node_profile.node_type != NodeType.FULL:
                continue
            
            if rid_type not in node_profile.provides.state:
                continue
            
            provider_nodes.append(node_rid)
        
        if provider_nodes:
            log.debug(f"Found provider(s) {provider_nodes}")
        else:
            log.debug("Failed to find providers")
            
        return provider_nodes
            
    def fetch_remote_bundle(self, rid: RID) -> tuple[Bundle | None, KoiNetNode | None]:
        """Attempts to fetch a bundle by RID from known peer nodes."""
        
        log.debug(f"Fetching remote bundle {rid!r}")
        remote_bundle, node_rid = None, None
        for node_rid in self.get_state_providers(type(rid)):
            try:
                payload = self.request_handler.fetch_bundles(
                    node=node_rid, rids=[rid])
            except RequestError:
                continue
            
            if payload.bundles:
                remote_bundle = payload.bundles[0]
                log.debug(f"Got bundle from {node_rid!r}")
                break
        
        if not remote_bundle:
            log.warning("Failed to fetch remote bundle")
            
        return remote_bundle, node_rid
    
    def fetch_remote_manifest(self, rid: RID) -> tuple[Bundle | None, KoiNetNode | None]:
        """Attempts to fetch a manifest by RID from known peer nodes."""
        
        log.debug(f"Fetching remote manifest {rid!r}")
        remote_manifest, node_rid = None, None
        for node_rid in self.get_state_providers(type(rid)):
            try:
                payload = self.request_handler.fetch_manifests(
                    node=node_rid, rids=[rid])
            except RequestError:
                continue
            
            if payload.manifests:
                remote_manifest = payload.manifests[0]
                log.debug(f"Got bundle from {node_rid!r}")
                break
        
        if not remote_manifest:
            log.warning("Failed to fetch remote bundle")
            
        return remote_manifest, node_rid
    
    def poll_neighbors(self) -> dict[KoiNetNode, list[Event]]:
        """Polls all neighbor nodes and returns compiled list of events.
        
        Neighbor nodes include any node this node shares an edge with,
        or the first contact, if no neighbors are found.
        
        NOTE: This function does not poll nodes that don't share edges
        with this node. Events sent by non neighboring nodes will not
        be polled.
        """
        
        neighbors: list[KoiNetNode] = []
        for node_rid in self.graph.get_neighbors():
            node_bundle = self.cache.read(node_rid)
            if not node_bundle: 
                continue
            node_profile = node_bundle.validate_contents(NodeProfile)
            if node_profile.node_type != NodeType.FULL: 
                continue
            neighbors.append(node_rid)
            
        if not neighbors and self.config.koi_net.first_contact.rid:
            neighbors.append(self.config.koi_net.first_contact.rid)
        
        event_dict: dict[KoiNetNode, list[Event]] = {}
        for node_rid in neighbors:
            try:
                payload = self.request_handler.poll_events(
                    node=node_rid, 
                    rid=self.identity.rid
                )
            except RequestError:
                continue
                
            log.debug(f"Received {len(payload.events)} events from {node_rid!r}")
            event_dict[node_rid] = payload.events
            
        return event_dict


import sys
import logging
from logging.handlers import RotatingFileHandler
from datetime import datetime
from typing import Callable

import structlog
import colorama


class LogSystem:
    """Configures and initializes the logging system."""

    shared_log_processors: list[Callable] = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.UnicodeDecoder(),
        structlog.processors.CallsiteParameterAdder({
            structlog.processors.CallsiteParameter.MODULE,
            structlog.processors.CallsiteParameter.FUNC_NAME
        }),
    ]
    
    use_file_handler: bool
    use_console_handler: bool
    file_handler_log_level: int
    console_handler_log_level: int
    
    log_file_path: str
    max_log_file_size: int
    num_log_file_backups: int
    log_file_encoding: str
    
    def __init__(
        self,
        use_file_handler: bool = True,
        use_console_handler: bool = True,
        file_handler_log_level: int = logging.DEBUG,
        console_handler_log_level: int = logging.INFO,
        log_file_path: str = "log.ndjson",
        max_log_file_size: int = 10 * 1024 ** 2,
        num_log_file_backups: int = 5,
        log_file_encoding: str = "utf-8"
    ):
        self.use_file_handler = use_file_handler
        self.use_console_handler = use_console_handler
        self.file_handler_log_level = file_handler_log_level
        self.console_handler_log_level = console_handler_log_level
        
        self.log_file_path = log_file_path
        self.max_log_file_size = max_log_file_size
        self.num_log_file_backups = num_log_file_backups
        self.log_file_encoding = log_file_encoding
        
        self.configure()
        
    def configure(self):
        handlers = []
        if self.use_file_handler:
            handlers.append(self.configure_file_handler())
        if self.use_console_handler:
            handlers.append(self.configure_console_handler())
        
        logging.basicConfig(level=logging.DEBUG, handlers=handlers)
        structlog.configure(
            processors=self.shared_log_processors + [
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
            wrapper_class=structlog.stdlib.BoundLogger,
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=True,
        )
        
    def configure_file_handler(self):
        file_handler = RotatingFileHandler(
            filename=self.log_file_path,
            maxBytes=self.max_log_file_size,
            backupCount=self.num_log_file_backups,
            encoding=self.log_file_encoding,
            delay=True
        )
        
        file_handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                processor=structlog.processors.JSONRenderer(),
                foreign_pre_chain=self.shared_log_processors
            )
        )
        
        file_handler.setLevel(self.file_handler_log_level)
        return file_handler
    
    def configure_console_handler(self):
        console_renderer = structlog.dev.ConsoleRenderer(
            columns=[
                # Render the timestamp without the key name in yellow.
                structlog.dev.Column(
                    "timestamp",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Style.DIM,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=lambda t: datetime.fromisoformat(t).strftime("%Y-%m-%d %H:%M:%S"),
                    ),
                ),
                structlog.dev.Column(
                    "level",
                    structlog.dev.LogLevelColumnFormatter(
                        level_styles={
                            level: colorama.Style.BRIGHT + color
                            for level, color in {
                                "critical": colorama.Fore.RED,
                                "exception": colorama.Fore.RED,
                                "error": colorama.Fore.RED,
                                "warn": colorama.Fore.YELLOW,
                                "warning": colorama.Fore.YELLOW,
                                "info": colorama.Fore.GREEN,
                                "debug": colorama.Fore.GREEN,
                                "notset": colorama.Back.RED,
                            }.items()
                        },
                        reset_style=colorama.Style.RESET_ALL,
                        width=9
                    )
                ),
                # Render the event without the key name in bright magenta.
                
                # Default formatter for all keys not explicitly mentioned. The key is
                # cyan, the value is green.
                structlog.dev.Column(
                    "path",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Fore.MAGENTA,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                        width=30
                    ),
                ),
                structlog.dev.Column(
                    "event",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=None,
                        value_style=colorama.Fore.WHITE,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                        width=30
                    ),
                ),
                structlog.dev.Column(
                    "",
                    structlog.dev.KeyValueColumnFormatter(
                        key_style=colorama.Fore.BLUE,
                        value_style=colorama.Fore.GREEN,
                        reset_style=colorama.Style.RESET_ALL,
                        value_repr=str,
                    ),
                )
            ]
        )
        
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                processor=console_renderer,
                foreign_pre_chain=self.shared_log_processors
            )
        )
        
        console_handler.setLevel(self.console_handler_log_level)
        return console_handler


import structlog
from queue import Queue
from rid_lib.core import RID
from rid_lib.ext import Bundle, Manifest
from rid_lib.types import KoiNetNode
from ..protocol.event import Event, EventType
from .knowledge_object import KnowledgeObject

log = structlog.stdlib.get_logger()


class KobjQueue:
    """Queue for knowledge objects entering the processing pipeline."""
    q: Queue[KnowledgeObject]
    
    def __init__(self):
        self.q = Queue()
                
    def push(
        self,
        rid: RID | None = None,
        manifest: Manifest | None = None,
        bundle: Bundle | None = None,
        event: Event | None = None,
        kobj: KnowledgeObject | None = None,
        event_type: EventType | None = None,
        source: KoiNetNode | None = None
    ):
        """Pushes knowledge object to queue.
        
        Input may take the form of an RID, manifest, bundle, event, 
        or knowledge object (with an optional event type for RID, 
        manifest, or bundle objects). All objects will be normalized 
        to knowledge objects and queued.
        """
        
        if rid:
            _kobj = KnowledgeObject.from_rid(rid, event_type, source)
        elif manifest:
            _kobj = KnowledgeObject.from_manifest(manifest, event_type, source)
        elif bundle:
            _kobj = KnowledgeObject.from_bundle(bundle, event_type, source)
        elif event:
            _kobj = KnowledgeObject.from_event(event, source)
        elif kobj:
            _kobj = kobj
        else:
            raise ValueError("One of 'rid', 'manifest', 'bundle', 'event', or 'kobj' must be provided")
        
        self.q.put(_kobj)
        log.debug(f"Queued {_kobj!r}")



import time
import typer
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from dotenv import load_dotenv

from ..interfaces.network import NetworkInterface
from ..interfaces.node import MissingEnvVariablesError, NodeExistsError


load_dotenv()


app = typer.Typer()
console = Console()


@app.command()
def create(node_type: str, node_name: str | None = None):
    # if node_type not in list(map(lambda ep: ep.name, installed_nodes)):
    #     console.print(f"[bold red]Error:[/bold red] node type '{node_type}' doesn't exist")
    #     raise typer.Exit(code=1)
    
    node_name = node_name or node_type
    
    network = NetworkInterface()
    try:
        network.create_node(node_name, node_type)
        init(node_name)
    except NodeExistsError:
        console.print(f"[red]Node '{node_name}' already exists[/red]")
    
@app.command()
def init(node_name: str):
    network = NetworkInterface()
    
    if node_name not in network.nodes:
        console.print(f"[red]Node '{node_name}' doesn't exist[/red]")
        return
    
    try:
        node = network.nodes[node_name]
        node.init()
        node_rid = node.get_config().koi_net.node_rid
        console.print(f"Initialized node '{node_name}' as {node_rid}")
    except MissingEnvVariablesError as err:
        text = "\n".join([f"[bold red]{v}[/bold red]" for v in err.vars])
        panel = Panel.fit(
            text, 
            border_style="red",
            title="Cannot initialize node, missing the following enironment variables:")
        console.print(panel)
        console.print(f"Run [cyan]koi init {node_name}[/cyan] after setting")
    
@app.command()
def delete(name: str):
    network = NetworkInterface()
    network.delete_node(name)
    
    if network.config.first_contact == name:
        network.config.first_contact = None
        network.config_loader.save_to_yaml()
    
@app.command()
def start(name: str, verbose: bool = False):
    network = NetworkInterface()
    node = network.nodes[name]
    node.start(suppress_output=not verbose)
    try:
        while node.process.poll() is None:
            time.sleep(0.1)
    except KeyboardInterrupt:
        node.stop()
        node.process.wait()


from pydantic import BaseModel

from ..config.loader import ConfigLoader


class KoiNetworkConfig(BaseModel):
    first_contact: str | None = None
    nodes: dict[str, str] = {}

class NetworkConfigLoader(ConfigLoader):
    file_path: str = "koi-network-config.yaml"


import structlog
import cryptography.exceptions
from rid_lib.ext import Bundle, Cache
from rid_lib.ext.utils import sha256_hash
from rid_lib.types import KoiNetNode
from .identity import NodeIdentity
from .protocol.envelope import UnsignedEnvelope, SignedEnvelope
from .protocol.secure import PublicKey
from .protocol.api_models import ApiModels, EventsPayload
from .protocol.event import EventType
from .protocol.node import NodeProfile
from .protocol.secure import PrivateKey
from .exceptions import (
    UnknownNodeError,
    InvalidKeyError,
    InvalidSignatureError,
    InvalidTargetError
)
from .config.base import BaseNodeConfig

log = structlog.stdlib.get_logger()


class SecureManager:
    """Subsystem handling secure protocol logic."""
    identity: NodeIdentity
    cache: Cache
    config: BaseNodeConfig
    priv_key: PrivateKey
    
    def __init__(
        self, 
        identity: NodeIdentity, 
        cache: Cache,
        config: BaseNodeConfig
    ):
        self.identity = identity
        self.cache = cache
        self.config = config
        
    def start(self):
        self.load_priv_key()
        
    def load_priv_key(self) -> PrivateKey:
        """Loads private key from PEM file path in config."""
        
        # TODO: handle missing private key
        with open(self.config.koi_net.private_key_pem_path, "r") as f:
            priv_key_pem = f.read()
        
        self.priv_key = PrivateKey.from_pem(
            priv_key_pem=priv_key_pem,
            password=self.config.env.priv_key_password
        )
        
    def handle_unknown_node(self, envelope: SignedEnvelope) -> Bundle | None:
        """Attempts to find node profile in proided envelope.
        
        If an unknown node sends an envelope, it may still be able to be
        validated if that envelope contains their node profile. This is
        essential for allowing unknown nodes to handshake and introduce
        themselves. Only an `EventsPayload` contain a `NEW` event for a 
        node profile for the source node is permissible.
        """
        if type(envelope.payload) != EventsPayload:
            return None
            
        for event in envelope.payload.events:
            # must be NEW event for bundle of source node's profile
            if event.rid != envelope.source_node:
                continue
            if event.event_type != EventType.NEW:
                continue            
            
            return event.bundle
        return None
        
    def create_envelope(
        self, payload: ApiModels, target: KoiNetNode
    ) -> SignedEnvelope:
        """Returns signed envelope to target from provided payload."""
        return UnsignedEnvelope(
            payload=payload,
            source_node=self.identity.rid,
            target_node=target
        ).sign_with(self.priv_key)
        
    def validate_envelope(self, envelope: SignedEnvelope):
        """Validates signed envelope from another node."""
        
        node_bundle = (
            self.cache.read(envelope.source_node) or
            self.handle_unknown_node(envelope)
        )
        
        if not node_bundle:
            raise UnknownNodeError(f"Couldn't resolve {envelope.source_node}")
        
        node_profile = node_bundle.validate_contents(NodeProfile)
        
        # check that public key matches source node RID
        if envelope.source_node.hash != sha256_hash(node_profile.public_key):
            raise InvalidKeyError("Invalid public key on new node!")
        
        # check envelope signed by validated public key
        pub_key = PublicKey.from_der(node_profile.public_key)
        try:
            envelope.verify_with(pub_key)
        except cryptography.exceptions.InvalidSignature:
            raise InvalidSignatureError(f"Signature {envelope.signature} is invalid.")
        
        # check that this node is the target of the envelope
        if envelope.target_node != self.identity.rid:
            raise InvalidTargetError(f"Envelope target {envelope.target_node!r} is not me")




import queue
import traceback
import time
import structlog

from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode

from ..config.base import BaseNodeConfig
from ..network.event_queue import EventQueue
from ..network.request_handler import RequestHandler
from ..network.event_buffer import EventBuffer
from ..protocol.node import NodeProfile, NodeType
from ..exceptions import RequestError
from .base import ThreadWorker, STOP_WORKER

log = structlog.stdlib.get_logger()


class EventProcessingWorker(ThreadWorker):
    """Thread worker that processes the `event_queue`."""
    
    def __init__(
        self,
        config: BaseNodeConfig,
        cache: Cache,
        event_queue: EventQueue,
        request_handler: RequestHandler,
        poll_event_buf: EventBuffer,
        broadcast_event_buf: EventBuffer
    ):
        self.event_queue = event_queue
        self.request_handler = request_handler
        
        self.config = config
        self.cache = cache
        self.poll_event_buf = poll_event_buf
        self.broadcast_event_buf = broadcast_event_buf
        
        super().__init__()
        
    def flush_and_broadcast(self, target: KoiNetNode, force_flush: bool = False):
        """Broadcasts all events to target in event buffer."""
        
        # TODO: deal with automated retries when unreachable node's buffer is full
        try:
            with self.broadcast_event_buf.safe_flush(target, force_flush) as events:
                self.request_handler.broadcast_events(target, events=events)
        except RequestError:
            log.warning("Failed to reach target, event buffer reset")
            pass
        
    def stop(self):
        self.event_queue.q.put(STOP_WORKER)
        super().stop()
    
    def run(self):
        while True:
            try:
                item = self.event_queue.q.get(
                    timeout=self.config.koi_net.event_worker.queue_timeout)
                
                try:
                    if item is STOP_WORKER:
                        log.info(f"Received 'STOP_WORKER' signal, flushing all buffers...")
                        for target in list(self.broadcast_event_buf.buffers.keys()):
                            self.flush_and_broadcast(target, force_flush=True)
                        return
                    
                    log.info(f"Dequeued {item.event!r} -> {item.target!r}")
                    
                    # determines which buffer to push event to based on target node type
                    node_bundle = self.cache.read(item.target)
                    if node_bundle:
                        node_profile = node_bundle.validate_contents(NodeProfile)
                        
                        if node_profile.node_type == NodeType.FULL:
                            self.broadcast_event_buf.push(item.target, item.event)
                            
                        elif node_profile.node_type == NodeType.PARTIAL:
                            self.poll_event_buf.push(item.target, item.event)
                            continue
                        
                    elif item.target == self.config.koi_net.first_contact.rid:
                        self.broadcast_event_buf.push(item.target, item.event)
                        
                    else:
                        log.warning(f"Couldn't handle event {item.event!r} in queue, node {item.target!r} unknown to me")
                        continue
                    
                    buf_len = self.broadcast_event_buf.buf_len(item.target)
                    if buf_len > self.config.koi_net.event_worker.max_buf_len:
                        self.flush_and_broadcast(target)

                finally:
                    self.event_queue.q.task_done()

            except queue.Empty:
                # On timeout, check all buffers for max wait time
                for target in list(self.broadcast_event_buf.buffers):
                    start_time = self.broadcast_event_buf.start_time.get(target)
                    
                    if (start_time is None) or (self.broadcast_event_buf.buf_len(target) == 0):
                        continue
                    
                    now = time.time()
                    if (now - start_time) >= self.config.koi_net.event_worker.max_wait_time: 
                        self.flush_and_broadcast(target)
                        
            except Exception:
                traceback.print_exc()
                continue


import queue
import traceback
import structlog

from ..config.base import BaseNodeConfig
from ..processor.pipeline import KnowledgePipeline
from ..processor.kobj_queue import KobjQueue
from .base import ThreadWorker, STOP_WORKER

log = structlog.stdlib.get_logger()


class KnowledgeProcessingWorker(ThreadWorker):
    """Thread worker that processes the `kobj_queue`."""
    
    def __init__(
        self,
        config: BaseNodeConfig,
        kobj_queue: KobjQueue,
        pipeline: KnowledgePipeline
    ):
        self.config = config
        self.kobj_queue = kobj_queue
        self.pipeline = pipeline

        super().__init__()
        
    def stop(self):
        self.kobj_queue.q.put(STOP_WORKER)
        super().stop()
        
    def run(self):
        while True:
            try:
                item = self.kobj_queue.q.get(timeout=self.config.koi_net.kobj_worker.queue_timeout)
                try:
                    if item is STOP_WORKER:
                        log.info("Received 'STOP_WORKER' signal, shutting down...")
                        return
                    
                    log.info(f"Dequeued {item!r}")
                    
                    self.pipeline.process(item)
                finally:
                    self.kobj_queue.q.task_done()
                    
            except queue.Empty:
                pass
            
            except Exception:
                traceback.print_exc()
                continue



import contextlib
import functools
import importlib
import inspect
import os
import shutil
import signal
import subprocess
import sys
from typing import Generator
from contextlib import contextmanager

from pydantic import ValidationError

from koi_net.config.base import BaseNodeConfig
from koi_net.config.env_config import EnvConfig
from koi_net.core import BaseNode

from ..exceptions import MissingEnvVariablesError, NodeExistsError


"""
entry point name -> module name -> node class, run node

node name (directory)
node type alias (entrypoint: `coordinator`)
node type name (module: `koi_net_coordinator_node`)

node name -> node type name: (stored in koi net config)
"""

ENTRY_POINT_GROUP = "koi_net.node"



class NodeInterface:
    def __init__(self, name: str, module: str):
        self.name = name
        self.module = module
        self.process = None
    
    @staticmethod
    def in_directory(fn):
        @functools.wraps(fn)
        def wrapper(self: "NodeInterface", *args, **kwargs):
            with contextlib.chdir(self.name):
                return fn(self, *args, **kwargs)
        return wrapper
    
    @property
    def node_class(self) -> type[BaseNode]:
        core = importlib.import_module(f"{self.module}.core")

        for _, obj in inspect.getmembers(core):
            # only look at objects defined in the module
            if getattr(obj, "__module__", None) != core.__name__:
                continue
            
            # identified node class for the module
            if issubclass(obj, BaseNode):
                return obj
    
    @classmethod
    def create(cls, name: str, module: str):
        try:
            os.mkdir(name)
        except FileExistsError:
            raise NodeExistsError(f"Node of name '{name}' already exists")
        
        return cls(name, module)
    
    @in_directory
    def init(self):
        for field in self.node_class.config_schema.model_fields.values():
            field_type = field.annotation
            if issubclass(field_type, EnvConfig):
                try:
                    field_type()
                except ValidationError as exc:
                    vars = [
                        err["loc"][0].upper()
                        for err in exc.errors()
                        if err["type"] == "missing"
                    ]
                    raise MissingEnvVariablesError(
                        message="Missing required environment variables",
                        vars=vars
                    )
        
        self.node_class().config_loader.start()
    
    @in_directory
    def get_config(self) -> BaseNodeConfig:
        return self.node_class().config

    @contextmanager
    @in_directory
    def mutate_config(self) -> Generator[BaseNodeConfig, None, None]:
        node = self.node_class()
        yield node.config
        node.config_loader.save_to_yaml()
    
    @in_directory
    def wipe(self):
        self.node_class().cache.drop()
    
    def delete(self):
        if self.process and self.process.poll() is None:
            print("Can't delete node while it's running, stop it first.")
            return False
        shutil.rmtree(self.name)
        return True
    
    @in_directory
    def start(self, suppress_output: bool = True):
        self.process = subprocess.Popen(
            (sys.executable, "-m", self.module),
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL if suppress_output else None,
            stderr=subprocess.DEVNULL if suppress_output else None,
            text=True
        )
    
    def stop(self):
        self.process.stdin.write("STOP\n")
        self.process.stdin.flush()

if __name__ == "__main__":
    node = NodeInterface("coordinator", "koi_net_coordinator_node")
    node.init()


import os
import shutil
from rid_lib.core import RID, RIDType
from rid_lib.ext import Bundle
from rid_lib.ext.utils import b64_encode, b64_decode

from .config.base import BaseNodeConfig


class Cache:
    def __init__(self, config: BaseNodeConfig):
        self.config = config
        
    @property
    def directory_path(self):
        return self.config.koi_net.cache_directory_path
        
    def file_path_to(self, rid: RID) -> str:
        encoded_rid_str = b64_encode(str(rid))
        return f"{self.directory_path}/{encoded_rid_str}.json"

    def write(self, bundle: Bundle) -> Bundle:
        """Writes bundle to cache, returns a Bundle."""
        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
            
        with open(
            file=self.file_path_to(bundle.manifest.rid), 
            mode="w", 
            encoding="utf-8"
        ) as f:
            f.write(bundle.model_dump_json(indent=2))

        return bundle
    
    def exists(self, rid: RID) -> bool:
        return os.path.exists(
            self.file_path_to(rid)
        )

    def read(self, rid: RID) -> Bundle | None:
        """Reads and returns CacheEntry from RID cache."""
        try:
            with open(
                file=self.file_path_to(rid), 
                mode="r",
                encoding="utf-8"
            ) as f:
                return Bundle.model_validate_json(f.read())
        except FileNotFoundError:
            return None
    
    def list_rids(self, rid_types: list[RIDType] | None = None) -> list[RID]:
        if not os.path.exists(self.directory_path):
            return []
        
        rids = []
        for filename in os.listdir(self.directory_path):
            encoded_rid_str = filename.split(".")[0]
            rid_str = b64_decode(encoded_rid_str)
            rid = RID.from_string(rid_str)
            
            if not rid_types or type(rid) in rid_types:
                rids.append(rid)
            
        return rids
                
    def delete(self, rid: RID) -> None:
        """Deletes cache bundle."""
        try:
            os.remove(self.file_path_to(rid))
        except FileNotFoundError:
            return

    def drop(self) -> None:
        """Deletes all cache bundles."""
        try:
            shutil.rmtree(self.directory_path)
        except FileNotFoundError:
            return




from enum import StrEnum
from pydantic import BaseModel
from rid_lib import RIDType
from rid_lib.ext.bundle import Bundle
from rid_lib.ext.utils import sha256_hash
from rid_lib.types import KoiNetEdge, KoiNetNode


class EdgeStatus(StrEnum):
    PROPOSED = "PROPOSED"
    APPROVED = "APPROVED"
    
class EdgeType(StrEnum):
    WEBHOOK = "WEBHOOK"
    POLL = "POLL"

class EdgeProfile(BaseModel):
    source: KoiNetNode
    target: KoiNetNode
    edge_type: EdgeType
    status: EdgeStatus
    rid_types: list[RIDType]


def generate_edge_bundle(
    source: KoiNetNode,
    target: KoiNetNode,
    rid_types: list[RIDType],
    edge_type: EdgeType
) -> Bundle:
    """Returns edge bundle."""
    
    edge_rid = KoiNetEdge(sha256_hash(
        str(source) + str(target)
    ))
    
    edge_profile = EdgeProfile(
        source=source,
        target=target,
        rid_types=rid_types,
        edge_type=edge_type,
        status=EdgeStatus.PROPOSED
    )
    
    edge_bundle = Bundle.generate(
        edge_rid,
        edge_profile.model_dump()
    )
    
    return edge_bundle


from typing import Generic, TypeVar
from ruamel.yaml import YAML

from .proxy import ConfigProxy
from .base import BaseNodeConfig


T = TypeVar("T", bound=BaseNodeConfig)

class ConfigLoader(Generic[T]):
    """Loads node config from a YAML file, and proxies access to it."""
    
    file_path: str = "config.yaml"
    file_content: str
    
    schema: type[T]
    proxy: ConfigProxy[T]
    
    def __init__(
        self, 
        config_schema: type[T],
        config: ConfigProxy[T]
    ):
        self.schema = config_schema
        self.proxy = config
        
        # this is a special case to allow config state dependent components
        # to initialize without a "lazy initialization" approach, in general
        # components SHOULD NOT execute code in their init phase
        self.load_from_yaml()
        
    def start(self):
        self.save_to_yaml()
    
    def load_from_yaml(self):
        """Loads config from YAML file, or generates it if missing."""
        yaml = YAML()
        
        try:
            with open(self.file_path, "r") as f:
                self.file_content = f.read()
            config_data = yaml.load(self.file_content)
            config = self.schema.model_validate(config_data)
            self.proxy._set_delegate(config)
        
        except FileNotFoundError:
            config = self.schema()
            self.proxy._set_delegate(config)
        
    def save_to_yaml(self):
        """Saves config to YAML file."""
        yaml = YAML()
        
        with open(self.file_path, "w") as f:
            try:
                config = self.proxy._get_delegate()
                config_data = config.model_dump(mode="json")
                yaml.dump(config_data, f)
                
            except Exception:
                # rewrites original content if YAML dump fails
                if self.file_content:
                    f.seek(0)
                    f.truncate()
                    f.write(self.file_content)
                raise


from pydantic_settings import BaseSettings, SettingsConfigDict


class EnvConfig(BaseSettings):
    """Config for environment variables.
    
    Variables set in this config class will be validated against the
    environment that the code is executed in. Names are case insensitive,
    so `priv_key_password` would validate `PRIV_KEY_PASSWORD` in the
    environment.
    """
    
    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
        case_sensitive=False
    )
    
    priv_key_password: str